/*********************************************************************
  Hide for PGP:
  a steganographic program that hides files in BMP, WAV or VOC files.
  If the hidden files resemble random noise - e.g. files encrypted with
  Pretty Good Privacy by Phil Zimmermann and without any header - e.g.
  the header information removed by Stealth by Henry Hastur - the hidden
  information cannot be detected by any means.

  copyright 1996 by Heinz Repp

  This program has been developped with PGP and Stealth in mind so it
  supports the stream mode of Stealth: the file to be hidden or extracted
  must be provided or read through a pipe or stdin/stdout redirection.

  usage:
  hide4pgp [-{1|2|4|8}] [-d[=]] [-p[+]] [-v[-]] <hostfile> < <file-to-hide>
  hide4pgp -x [-{1|2|4|8}] [-v[-]] <hostfile> > <extracted-file>

  parameters may be any order and switches combined

*********************************************************************/

/* includes */
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <io.h>
#include <fcntl.h>
#include "hide4pgp.h"

/* globals */
FILE *StegFile;                              /* data file to work on */
int hiding = TRUE, pairing = 0, duplic = 0, verbose = 1;    /* flags */
int BitCount = 0;                                /* no. of LSBs used */
int BitMask;                                 /* to mask out the LSBs */
long capacity = 0, occupied = 0;  /* char counters from/to stdin/out */
size_t data_size;                             /* data width in bytes */
int (*get_data) (FILE *);  /* funct pointer: read one item from file */
void (*mod_data) (void *, size_t);   /* funct pointer: modify buffer */

/* extern declarations */
extern int bmp_prepare (void);
extern int wav_prepare (void);
extern int voc_prepare (void);
extern long bmp_nextblk (void);
extern long wav_nextblk (void);
extern long voc_nextblk (void);


/*********************************************************************
  set_bits:  sets the least significant bits of a byte/word according
             to the next bits of the secret data; keeps a byte wide
             buffer for the secret data and fills it by reading from
             stdin when empty; after EOF read the buffer is filled
             with random data; the global variables capacity and occu-
             pied are set according to total bytes and read up to EOF
  parameter: the (8 or 16 bit) value in whose least significant bits
             the next bits of the secret data are to be stored
  returns:   an int value with the corresponding bits set
*********************************************************************/
int set_bits (int original_data)
{ static int byte_in_work;
  static int count = 0;
  static int from_file = TRUE;

  if (count == 0)
  { if (from_file)
    { if ((byte_in_work = getchar ()) == EOF)
      { from_file = FALSE;
        srand ((unsigned int) clock ());
        byte_in_work = rand ();
        occupied = capacity;
      } /* endif EOF */
    } else
    { byte_in_work = rand ();
    } /* endif fromfile */
    count = 8;
    capacity++;
  } /* endif count == 0 */
  return byte_in_work >> (count -= BitCount)  & BitMask |
         original_data & ~BitMask;
} /* set_bits */


/*********************************************************************
  get_bits:  extracts the hidden bits out of previously modified data
             and rearranges them to bytes in a local shift register
             that is emptied to stdout when full; the number of bytes
             output is recorded to capacity
  parameter: value with secret data in its least significant bits
  returns:   nothing
*********************************************************************/
void get_bits (int hiding_data)
{ static int byte_in_work;
  static int count = 0;

  byte_in_work <<= BitCount;
  byte_in_work |= hiding_data & BitMask;
  if ((count += BitCount) == 8)
  { putchar (byte_in_work);
    count = 0;
    capacity++;
  } /* endif bitcount == 8 */
} /* get_bits *


/*********************************************************************
  unexp_EOF: is called when read past EOF, outputs an error message
             and aborts
  parameter: none
  returns:   never
*********************************************************************/
void unexp_EOF (void)
{ fputs ("\nError: Unexpected end of file encountered - aborting.\n",
         stderr);
  exit (9);
} /* unexp_EOF */


/*********************************************************************
  no_mem:    is called when malloc fails, outputs an error message
             and aborts
  parameter: none
  returns:   never
*********************************************************************/
void no_mem (void)
{ fputs ("\nError: Not enough memory - aborting.\n",
         stderr);
  exit (8);
} /* unexp_EOF */


/*********************************************************************
  mod_bytes: injects the secret data bits into a block of 8 bit data
             in memory
  parameter: pointer to char-array, size of array
  returns:   nothing
*********************************************************************/
void mod_bytes (char *block, size_t count)
{ while (count--)
  { *block = set_bits (*block);
    block++;
  } /* while count */
} /* mod_bytes */


/*********************************************************************
  mod_words: injects the secret data bits into a block of 16 bit data
             in memory
  parameter: pointer to int-array, size of array
  returns:   nothing
*********************************************************************/
void mod_words (int *block, size_t count)
{ while (count--)
  { *block = set_bits (*block);
    block++;
  } /* while count */
} /* mod_words */


/*********************************************************************
  set_block: set the secret bits in a block of the disk file; tries to
             read the block into memory; if not successful, splits the
             block into halves and calls itself recursively
  parameter: length of the continous block in bytes, file pointer set
             to first byte
  returns:   nothing
*********************************************************************/
void set_block (long length)
{ void *block;
  long position;
  size_t count;

  if (length == (count = length) &&
      (length ? (block = malloc (count)) : (no_mem (), NULL)))
       /* if length == 0 no_mem aborts and never returns */
  { count /= data_size;
    position = ftell (StegFile);
    fread (block, data_size, count, StegFile);
    if (feof (StegFile)) unexp_EOF;
    (*mod_data) (block, count);
    fseek (StegFile, position, SEEK_SET);
    fwrite (block, data_size, count, StegFile);
    free (block);
  } else
  { /* length > size_t or malloc failed */
    set_block (length++ >> 1);
    fseek (StegFile, 0L, SEEK_CUR);
    set_block (length >> 1);
  } /* endif block size */
} /* set_block */


/*********************************************************************
  get_block: get bits from a block of the disk file
  parameter: length of the continous block, read pointer set to first
             byte
  returns:   nothing
*********************************************************************/
void get_block (long length)
{ length /= data_size;
  do
  { get_bits ((*get_data) (StegFile));
    if (feof (StegFile)) unexp_EOF;
  } while (--length);
} /* get_block */


/*********************************************************************
  next_is:   compare string to next bytes in file, allways read the
             whole length
  parameter: pointer to a string to be matched
  returns:   flag, TRUE if equal
             file pointer set behind the characters compared
*********************************************************************/
int next_is (char *pattern)
{ int c, result = TRUE;
  while (*pattern)
  { if ((c = fgetc (StegFile)) == EOF) unexp_EOF;
    if (c != *pattern++) result = FALSE;
  } /* while length */
  return result;
} /* next_is */


/*********************************************************************
  hello:  show initial prompt and copyright
*********************************************************************/
void hello (void)
{ fputs ("Hide4PGP v1.0 - hides secret files in BMP/WAV/VOC\n"
         "(c) 1996 by Heinz Repp\n\n", stderr);
} /* hello */


/*********************************************************************
  usage:  show syntax
*********************************************************************/
void usage (void)
{ hello ();
  fputs (
"usage:\n"
"    Hide4PGP [-{1|2|4|8}] [-d[=]] [-p[+]] [-v[-]] <datafile> < <file-to-hide>\n"
"or: Hide4PGP -x [-{1|2|4|8}] [-v[-]] <datafile> > <extracted-file>\n"
"\n"
"    Hide4PGP -h for help.\n", stderr);
} /* usage */


/*********************************************************************
  help:  show detailed help
*********************************************************************/
void help (void)
{ hello ();
  fputs (
"To hide secret data in a data file:\n"
"  Hide4PGP [-option] <datafile> < <secret.in>          or using pipes:\n"
"  pgp -ef <YourID> < <secret.in> | stealth | Hide4PGP [-option] <datafile>\n"
"\n"
"To extract secret data out of a data file:\n"
"  Hide4PGP -x [-option] <datafile> > <secret.out>      or using pipes:\n"
"  Hide4PGP -x [-option] datafile | stealth -a <MyID> | pgp -f > <secret.out>\n"
"\n"
"  (!)  Secret data are always read from and written to standard in/out,\n"
"                 so the secret file MUST use i/o-redirection or a pipe!\n"
"\n"
"options may be in any order or combined:\n"
" x        eXtract (see above, default action is hiding)\n"
" 1|2|4|8  number of least significant bits used (default = 1 with 8 bit and\n"
"          = 4 with 16 bit data); 8 is only allowed with 16 bit data\n"
" d[=]     Duplicate used palette entries (BMP) d=: treat all entries equal\n"
" p[+]     Pair similar palette entries (BMP)   p+: use alternate method\n"
" v[-]     Verbose: give detailed information   v-: only error messages\n"
" h        Help (this screen)\n", stderr);
} /* help */


/*********************************************************************
  main:       parses the command line, reads the first byte in the
              StegFile, decides which file type to assume, and then
              does most of the work by vectorized calls to type and
              data width specific routines
  parameters: one file name (required) and optional switches
*********************************************************************/
int main (int argc, char *argv[])
{ int (*i_prepare) (void);
  long (*i_nextblk) (void);
  void (*i_editblk) (long);
  int steg_pos = 0;
  long block;
  char swch;

  if (! --argc)
  { usage ();
    exit (1);
  } /* endif */

  /* parse command line */
  do
  { switch (*argv[argc])
    { case '-':
      case '/':
        while (swch = *++argv[argc])
        { switch (swch)
          { case '1':
            case '2':
            case '4':
            case '8':
              BitCount = swch - '0';
              break;
            case 'H':
            case 'h':
            case '?':
              help ();
              exit (0);
            case 'D':
            case 'd':
              if (argv[argc][1] == '=')
              { duplic = 2;
                ++argv[argc];
              } else duplic = 1;
              break;
            case 'P':
            case 'p':
              if (argv[argc][1] == '+')
              { pairing = 2;
                ++argv[argc];
              } else pairing = 1;
              break;
            case 'V':
            case 'v':
              if (argv[argc][1] == '-')
              { verbose = 0;
                ++argv[argc];
              } else verbose = 2;
              break;
            case 'X':
            case 'x':
              hiding = FALSE;
              break;
            default:
              if (verbose) hello ();
              fputs ("Error: Invalid switch.\n", stderr);
              exit (2);
          } /* switch swch */
        } /* while swch */
        break;
      default:
        if (steg_pos)
        { if (verbose) hello ();
          fputs ("Error: Only ONE filename allowed as parameter.\n", stderr);
          exit (3);
        } else
        { steg_pos = argc;
        } /* endif */
    } /* switch commandline parameter */
  } while (--argc);

  if (verbose) hello ();
  if (! steg_pos)
  { fputs ("Error: Required filename missing.\n", stderr);
    exit (4);
  } /* endif */

  if ((StegFile = fopen (argv[steg_pos], hiding ? "r+b" : "rb")) == NULL)
  { fprintf (stderr, "Error: Cannot open stegfile %s.\n", argv[steg_pos]);
    exit (5);
  } /* endif */

  /* determine data file type from first char in file */
  switch (fgetc (StegFile))
  { case 'B':
      i_prepare = bmp_prepare;
      i_nextblk = bmp_nextblk;
      break;
    case 'C':
      i_prepare = voc_prepare;
      i_nextblk = voc_nextblk;
      break;
    case 'R':
      i_prepare = wav_prepare;
      i_nextblk = wav_nextblk;
      break;
    default:
      fputs ("Error: File format not recognized.\n", stderr);
      exit (6);
  } /* switch */

  /* check and initialize data file, set vectorized data routines */
  if ((*i_prepare) ())
  { data_size = sizeof (int);
    get_data = GETWORD;
    mod_data = mod_words;
    if (! BitCount) BitCount = 4;
  } else
  { data_size = sizeof (char);
    get_data = fgetc;
    mod_data = mod_bytes;
    switch (BitCount)
    { case 8:
        fputs ("Error: 8 bit only useful with 16 bit data.\n", stderr);
        exit (7);
      case 0:
        BitCount = 1;
    } /* switch */
  } /* endif 16 bit */
  BitMask = ~(~0 << BitCount);

  /* set stdin/stdout to binary - change if setmode not available */
  setmode (fileno (hiding ? stdin : stdout), O_BINARY);
  i_editblk = hiding ? set_block : get_block;

  /* main workhorse loop */
  while (block = (*i_nextblk) ())
    (*i_editblk) (block);

  /* clean up and summarize */
  fclose (StegFile);
  if (hiding)
  { set_bits (0); /* dummy */
    if (occupied)
    { if (verbose > 1)
        fprintf (stderr,
                 "Encrypted file uses %li out of %li available bytes.\n",
                 occupied, --capacity);
    } else
    { if (verbose)
        fprintf (stderr,
                 "*** Warning: Encrypted file truncated at %li Bytes"
           " ***\n             will probably not decode completely!!\n",
                 --capacity);
    } /* endif occupied */
  } else
  { if (verbose > 1)
      fprintf (stderr, "%li bytes extracted - encrypted file may be "
               "considerably smaller!\n", capacity);
  } /* endif hiding */
  if (verbose)
    fputs ("Steganographic operation finished.\n", stderr);
  return 0;
} /* main */
